# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Facilities for converting JSON <-> Python objects"""

from __future__ import absolute_import

import collections
import copy
import inspect
import numbers
import threading

import numpy as np
import six

from six.moves import range

from .json_utils import encode_json_for_repr


def to_json(value):
    if isinstance(value, np.ndarray):
        return value.tolist()
    try:
        method = value.to_json
    except AttributeError:
        return value
    return method()


class JsonObjectWrapper(object):
    supports_readonly = True

    __slots__ = ('_json_data', '_cached_wrappers', '_lock', '_readonly')

    def __init__(self, json_data=None, _readonly=False, **kwargs):
        if json_data is None:
            json_data = collections.OrderedDict()
        elif isinstance(json_data, type(self)):
            json_data = json_data.to_json()
        elif not isinstance(json_data, dict):
            raise TypeError
        object.__setattr__(self, '_json_data', json_data)
        object.__setattr__(self, '_cached_wrappers', dict())
        object.__setattr__(self, '_lock', threading.RLock())
        object.__setattr__(self, '_readonly', 1 if _readonly else False)
        for k in kwargs:
            setattr(self, k, kwargs[k])
        object.__setattr__(self, '_readonly', _readonly)

    def to_json(self):
        if self._readonly:
            return self._json_data
        with self._lock:
            r = self._json_data.copy()
            for k, (wrapper, _) in six.iteritems(self._cached_wrappers):
                if wrapper is not None:
                    r[k] = to_json(wrapper)
                else:
                    r.pop(k, None)
            return r

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(self.to_json(), memo))

    def __eq__(self, other):
        return type(self) == type(other) and self.to_json() == other.to_json()

    def __repr__(self):
        return u'%s(%s)' % (type(self).__name__, encode_json_for_repr(self.to_json()))

    def _get_wrapped(self, key, wrapped_type):
        with self._lock:
            json_value = self._json_data.get(key)
            cached_value = self._cached_wrappers.get(key)
            if cached_value is not None and cached_value[1] is json_value:
                return cached_value[0]
            kwargs = dict()
            if self._readonly and hasattr(wrapped_type, 'supports_readonly'):
                kwargs['_readonly'] = True
            wrapper = wrapped_type(json_value, **kwargs)
            self._cached_wrappers[key] = wrapper, json_value
            return wrapper

    def _set_wrapped(self, key, value, validator):
        if self._readonly is True:
            raise AttributeError
        value = validator(value)
        with self._lock:
            self._cached_wrappers[key] = (value, self._json_data.get(key))

_types_supporting_validation = frozenset([np.uint64])

def _normalize_validator(wrapped_type, validator):
    if validator is None:
        if (inspect.isroutine(wrapped_type) or hasattr(wrapped_type, 'supports_validation')
                or wrapped_type in _types_supporting_validation):
            validator = wrapped_type
        else:
            def validator_func(x):
                if not isinstance(x, wrapped_type):
                    raise TypeError
                return x
            validator = validator_func
    return validator


def wrapped_property(json_name, wrapped_type, validator=None, doc=None):
    validator = _normalize_validator(wrapped_type, validator)
    return property(fget=lambda self: self._get_wrapped(json_name, wrapped_type),
                    fset=lambda self, value: self._set_wrapped(json_name, value, validator),
                    doc=doc)


def array_wrapper(dtype, shape=None):
    if shape is not None:
        if isinstance(shape, numbers.Number):
            shape = (shape, )
        else:
            shape = tuple(shape)

    def wrapper(value, _readonly=False):
        value = np.array(value, dtype=dtype)
        if _readonly:
            value.setflags(write=False)
        if shape is not None:
            if len(shape) != len(value.shape) or any(
                    expected_size is not None and expected_size != actual_size
                    for expected_size, actual_size in zip(shape, value.shape)):
                raise ValueError('expected shape', shape)
        return value

    wrapper.supports_readonly = True
    return wrapper


def text_type(value):
    return six.text_type(value)


def optional(wrapper, default_value=None):
    def modified_wrapper(value, **kwargs):
        if value is None:
            return default_value
        return wrapper(value, **kwargs)

    if hasattr(wrapper, 'supports_readonly'):
        modified_wrapper.supports_readonly = True
    return modified_wrapper

class MapBase(object):
    pass

def typed_string_map(wrapped_type, validator=None):
    validator = _normalize_validator(wrapped_type, validator)
    class Map(JsonObjectWrapper, MapBase):
        supports_validation = True

        def __init__(self, json_data=None, _readonly=False):
            if isinstance(json_data, MapBase):
                json_data = json_data.to_json()
            elif json_data is not None:
                new_map = collections.OrderedDict()
                for k, v in six.viewitems(json_data):
                    validator(v)
                    new_map[k] = to_json(v)
                json_data = new_map
            super(Map, self).__init__(json_data, _readonly=_readonly)

        def clear(self):
            with self._lock:
                self._cached_wrappers.clear()
                self._json_data.clear()

        def keys(self):
            return six.viewkeys(self._json_data)

        def iteritems(self):
            for key in self:
                yield (key, self[key])

        def itervalues(self):
            for key in self:
                yield self[key]

        def get(self, key, default_value=None):
            with self._lock:
                if key in self._json_data:
                    return self[key]
                return default_value

        def __len__(self):
            return len(self._json_data)

        def __contains__(self, key):
            return key in self._json_data

        def __getitem__(self, key):
            with self._lock:
                if key not in self._json_data:
                    raise KeyError
                return self._get_wrapped(key, wrapped_type)

        def __setitem__(self, key, value):
            with self._lock:
                self._set_wrapped(key, value, validator)
                self._json_data[key] = None # placeholder

        def __delitem__(self, key):
            if self._readonly:
                raise AttributeError
            with self._lock:
                del self._json_data[key]
                self._cached_wrappers.pop(key, None)

        def __iter__(self):
            return iter(self._json_data)

    return Map

def typed_set(wrapped_type):
    def wrapper(x, _readonly=False):
        set_type = frozenset if _readonly else set
        kwargs = dict()
        if hasattr(wrapped_type, 'supports_readonly'):
            kwargs.update(_readonly=True)
        if x is None:
            return set_type()
        return set_type(wrapped_type(v, **kwargs) for v in x)
    wrapper.supports_readonly = True
    return wrapper

def typed_list(wrapped_type, validator=None):
    validator = _normalize_validator(wrapped_type, validator)
    class TypedList(object):
        supports_readonly = True
        supports_validation = True
        def __init__(self, json_data=None, _readonly=False):
            if json_data is None:
                json_data = []
            if not isinstance(json_data, (list, tuple, np.ndarray)):
                raise ValueError
            self._readonly = _readonly
            self._data = [validator(x) for x in json_data]

        def __len__(self):
            return len(self._data)

        def __getitem__(self, key):
            return self._data[key]

        def __delitem__(self, key):
            if self._readonly:
                raise AttributeError
            del self._data[key]

        def __setitem__(self, key, value):
            if self._readonly:
                raise AttributeError
            if isinstance(key, slice):
                values = [validator(x) for x in value]
                self._data[key] = values
            else:
                value = validator(x)
                self._data[key] = value

        def __iter__(self):
            return iter(self._data)

        def append(self, x):
            if self._readonly:
                raise AttributeError
            x = validator(x)
            self._data.append(x)

        def extend(self, values):
            for x in values:
                self.append(x)
        def insert(self, index, x):
            x = validator(x)
            self._data.insert(index, x)

        def pop(self, index=-1):
            return self._data.pop(index)

        def to_json(self):
            return [to_json(x) for x in self._data]

        def __deepcopy__(self, memo):
            return type(self)(copy.deepcopy(self.to_json(), memo))

        def __repr__(self):
            return encode_json_for_repr(self.to_json())
    return TypedList



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import contextlib
import json
import re
import threading

import six

from . import local_volume, trackable_state, viewer_config_state, viewer_state
from . import skeleton
from .json_utils import decode_json, encode_json, json_encoder_default
from .random_token import make_random_token


class LocalVolumeManager(trackable_state.ChangeNotifier):
    def __init__(self, token_prefix):
        super(LocalVolumeManager, self).__init__()
        self.volumes = dict()
        self.__token_prefix = token_prefix

    def register_volume(self, v):
        if v.token not in self.volumes:
            self.volumes[v.token] = v
            self._dispatch_changed_callbacks()
        if isinstance(v, local_volume.LocalVolume):
            source_type = 'volume'
        else:
            source_type = 'skeleton'
        return 'python://%s/%s' % (source_type, self.get_volume_key(v))

    def get_volume_key(self, v):
        return self.__token_prefix + v.token

    def update(self, json_str):
        pattern = '|'.join(self.volumes)
        present_tokens = set()
        for m in re.finditer(pattern, json_str):
            present_tokens.add(m.group(0))
        volumes_to_delete = []
        for x in self.volumes:
            if x not in present_tokens:
                volumes_to_delete.append(x)
        for x in volumes_to_delete:
            del self.volumes[x]
        if volumes_to_delete:
            self._dispatch_changed_callbacks()


class ViewerCommonBase(object):
    def __init__(self):
        self.token = make_random_token()
        self.config_state = trackable_state.TrackableState(viewer_config_state.ConfigState)

        def set_actions(actions):
            def func(s):
                s.actions = actions

            self.config_state.retry_txn(func)

        self.actions = viewer_config_state.Actions(set_actions)

        self.volume_manager = LocalVolumeManager(self.token + '.')

        self.__watched_volumes = dict()

        self.volume_manager.add_changed_callback(self._handle_volumes_changed)

        self._next_screenshot_id = 0
        self._screenshot_callbacks = {}
        self.actions.add('screenshot', self._handle_screenshot_reply)

    def async_screenshot(self, callback):
        screenshot_id = str(self._next_screenshot_id)
        self._next_screenshot_id += 1
        def set_screenshot_id(s):
            s.screenshot = screenshot_id
        self.config_state.retry_txn(set_screenshot_id)
        self._screenshot_callbacks[screenshot_id] = callback

    def screenshot(self, size=None):
        """Capture a screenshot synchronously.

        :param size: Optional.  List of [width, height] specifying the dimension
                     in pixels of the canvas to use.  If specified, UI controls
                     are hidden and the canvas is resized to the specified
                     dimensions while the screenshot is captured.

        :returns: The screenshot.
        """
        if size is not None:
            prior_state = self.config_state.state
            with self.config_state.txn() as s:
                s.show_ui_controls = False
                s.show_panel_borders = False
                s.viewer_size = size
        event = threading.Event()
        result = [None]
        def handler(s):
            result[0] = s
            event.set()
        self.async_screenshot(handler)
        event.wait()
        if size is not None:
            self.config_state.set_state(prior_state)
        return result[0]

    def _handle_screenshot_reply(self, s):
        def set_screenshot_id(s):
            s.screenshot = None

        self.config_state.retry_txn(set_screenshot_id)
        screenshot_id = s.screenshot.id
        callback = self._screenshot_callbacks.pop(screenshot_id, None)
        if callback is not None:
            callback(s)

    def _handle_volumes_changed(self):
        volumes = self.volume_manager.volumes
        for key in volumes:
            if key not in self.__watched_volumes:
                volume = volumes[key]
                self.__watched_volumes[key] = volume
                volume.add_changed_callback(self._update_source_generations)
        keys_to_remove = [key for key in self.__watched_volumes if key not in volumes]
        for key in keys_to_remove:
            volume = self.__watched_volumes.pop(key)
            volume.remove_changed_callback(self._update_source_generations)

    def _update_source_generations(self):
        def func(s):
            volume_manager = self.volume_manager
            s.source_generations = {
                volume_manager.get_volume_key(x): x.change_count for x in six.viewvalues(self.volume_manager.volumes)
            }

        self.config_state.retry_txn(func)

    def _transform_viewer_state(self, new_state):
        if isinstance(new_state, viewer_state.ViewerState):
            new_state = new_state.to_json()

            def encoder(x):
                if isinstance(x, (local_volume.LocalVolume, skeleton.SkeletonSource)):
                    return self.volume_manager.register_volume(x)
                return json_encoder_default(x)

            new_state = decode_json(json.dumps(new_state, default=encoder))
        return new_state

    def txn(self):
        raise NotImplementedError


class ViewerBase(ViewerCommonBase):
    def __init__(self):
        super(ViewerBase, self).__init__()
        self.shared_state = trackable_state.TrackableState(viewer_state.ViewerState,
                                                           self._transform_viewer_state)
        self.shared_state.add_changed_callback(
            lambda: self.volume_manager.update(encode_json(self.shared_state.raw_state)))

    @property
    def state(self):
        return self.shared_state.state

    def set_state(self, *args, **kwargs):
        return self.shared_state.set_state(*args, **kwargs)

    def txn(self, *args, **kwargs):
        return self.shared_state.txn(*args, **kwargs)

    def retry_txn(self, *args, **kwargs):
        return self.shared_state.retry_txn(*args, **kwargs)


class UnsynchronizedViewerBase(ViewerCommonBase):
    def __init__(self):
        super(UnsynchronizedViewerBase, self).__init__()
        self.state = viewer_state.ViewerState()

    @property
    def raw_state(self):
        return self._transform_viewer_state(self.state)

    def set_state(self, new_state):
        self.state = viewer_state.ViewerState(new_state)

    @contextlib.contextmanager
    def txn(self):
        yield self.state

    def retry_txn(self, func, retries=None):  # pylint: disable=unused-argument
        return func(self.state)



import os

def check_golden_contents(path, expected_contents, write=None):
    if write is None:
        write = os.getenv('NEUROGLANCER_GENERATE_GOLDEN') == '1'
    if write:
        with open(path, 'wb') as f:
            f.write(expected_contents)
    else:
        with open(path, 'rb') as f:
            contents = f.read()
        assert contents == expected_contents



from __future__ import division

import json
import os
import copy
import collections
import argparse
import csv

import neuroglancer
import numpy as np


class State(object):
    def __init__(self, path):
        self.path = path
        self.body_labels = collections.OrderedDict()

    def load(self):
        if os.path.exists(self.path):
            with open(self.path, 'r') as f:
                self.body_labels = collections.OrderedDict(json.load(f))

    def save(self):
        tmp_path = self.path + '.tmp'
        with open(tmp_path, 'w') as f:
            f.write(json.dumps(self.body_labels.items()))
        os.rename(tmp_path, self.path)


Body = collections.namedtuple('Body', ['segment_id', 'num_voxels', 'bbox_start', 'bbox_size'])


class Tool(object):
    def __init__(self, state_path, bodies, labels, segmentation_url, image_url, num_to_prefetch):
        self.state = State(state_path)
        self.num_to_prefetch = num_to_prefetch
        self.viewer = neuroglancer.Viewer()
        self.bodies = bodies
        self.state.load()
        self.total_voxels = sum(x.num_voxels for x in bodies)
        self.cumulative_voxels = np.cumsum([x.num_voxels for x in bodies])

        with self.viewer.txn() as s:
            s.layers['image'] = neuroglancer.ImageLayer(source=image_url)
            s.layers['segmentation'] = neuroglancer.SegmentationLayer(source=segmentation_url)
            s.show_slices = False
            s.concurrent_downloads = 256
            s.gpu_memory_limit = 2 * 1024 * 1024 * 1024
            s.layout = '3d'

        key_bindings = [
            ['bracketleft', 'prev-index'],
            ['bracketright', 'next-index'],
            ['home', 'first-index'],
            ['end', 'last-index'],
            ['control+keys', 'save'],
        ]
        label_keys = ['keyd', 'keyf', 'keyg', 'keyh']
        for label, label_key in zip(labels, label_keys):
            key_bindings.append([label_key, 'label-%s' % label])

            def label_func(s, label=label):
                self.set_label(s, label)

            self.viewer.actions.add('label-%s' % label, label_func)
        self.viewer.actions.add('prev-index', self._prev_index)
        self.viewer.actions.add('next-index', self._next_index)
        self.viewer.actions.add('first-index', self._first_index)
        self.viewer.actions.add('last-index', self._last_index)
        self.viewer.actions.add('save', self.save)

        with self.viewer.config_state.txn() as s:
            for key, command in key_bindings:
                s.input_event_bindings.viewer[key] = command
            s.status_messages['help'] = ('KEYS: ' + ' | '.join('%s=%s' % (key, command)
                                                               for key, command in key_bindings))

        self.index = -1
        self.set_index(self._find_one_after_last_labeled_index())

    def _find_one_after_last_labeled_index(self):
        body_index = 0
        while self.bodies[body_index].segment_id in self.state.body_labels:
            body_index += 1
        return body_index

    def set_index(self, index):
        if index == self.index:
            return
        body = self.bodies[index]
        self.index = index

        def modify_state_for_body(s, body):
            s.layers['segmentation'].segments = frozenset([body.segment_id])
            s.voxel_coordinates = body.bbox_start + body.bbox_size // 2

        with self.viewer.txn() as s:
            modify_state_for_body(s, body)

        prefetch_states = []
        for i in range(self.num_to_prefetch):
            prefetch_index = self.index + i + 1
            if prefetch_index >= len(self.bodies):
                break
            prefetch_state = copy.deepcopy(self.viewer.state)
            prefetch_state.layout = '3d'
            modify_state_for_body(prefetch_state, self.bodies[prefetch_index])
            prefetch_states.append(prefetch_state)

        with self.viewer.config_state.txn() as s:
            s.prefetch = [
                neuroglancer.PrefetchState(state=prefetch_state, priority=-i)
                for i, prefetch_state in enumerate(prefetch_states)
            ]

        label = self.state.body_labels.get(body.segment_id, '')
        with self.viewer.config_state.txn() as s:
            s.status_messages['status'] = (
                '[Segment %d/%d  : %d/%d voxels labeled = %.3f fraction] label=%s' %
                (index, len(self.bodies), self.cumulative_voxels[index], self.total_voxels,
                 self.cumulative_voxels[index] / self.total_voxels, label))

    def save(self, s):
        self.state.save()

    def set_label(self, s, label):
        self.state.body_labels[self.bodies[self.index].segment_id] = label
        self.set_index(self.index + 1)

    def _first_index(self, s):
        self.set_index(0)

    def _last_index(self, s):
        self.set_index(max(0, self._find_one_after_last_labeled_index() - 1))

    def _next_index(self, s):
        self.set_index(self.index + 1)

    def _prev_index(self, s):
        self.set_index(max(0, self.index - 1))


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument('--image-url', required=True, help='Neuroglancer data source URL for image')
    ap.add_argument(
        '--segmentation-url', required=True, help='Neuroglancer data source URL for segmentation')

    ap.add_argument('--state', required=True, help='Path to proofreading state file')

    ap.add_argument('--bodies', required=True, help='Path to list of bodies to proofread')
    ap.add_argument(
        '-a',
        '--bind-address',
        help='Bind address for Python web server.  Use 127.0.0.1 (the default) to restrict access '
        'to browers running on the local machine, use 0.0.0.0 to permit access from remote browsers.'
    )
    ap.add_argument(
        '--static-content-url', help='Obtain the Neuroglancer client code from the specified URL.')

    ap.add_argument('--labels', nargs='+', help='Labels to use')
    ap.add_argument('--prefetch', type=int, default=10, help='Number of bodies to prefetch')

    args = ap.parse_args()
    if args.bind_address:
        neuroglancer.set_server_bind_address(args.bind_address)
    if args.static_content_url:
        neuroglancer.set_static_content_source(url=args.static_content_url)

    bodies = []

    with open(args.bodies, 'r') as f:
        csv_reader = csv.DictReader(f)
        for row in csv_reader:
            bodies.append(
                Body(
                    segment_id=int(row['id']),
                    num_voxels=int(row['num_voxels']),
                    bbox_start=np.array(
                        [
                            int(row['bbox.start.x']),
                            int(row['bbox.start.y']),
                            int(row['bbox.start.z'])
                        ],
                        dtype=np.int64),
                    bbox_size=np.array(
                        [int(row['bbox.size.x']),
                         int(row['bbox.size.y']),
                         int(row['bbox.size.z'])],
                        dtype=np.int64),
                ))

    tool = Tool(
        state_path=args.state,
        image_url=args.image_url,
        segmentation_url=args.segmentation_url,
        labels=args.labels,
        bodies=bodies,
        num_to_prefetch=args.prefetch,
    )
    print(tool.viewer)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import re

import six

import sockjs.tornado

from . import trackable_state, viewer_config_state
from .json_utils import decode_json, encode_json

SOCKET_PATH_REGEX_WITHOUT_GROUP = r'^/socket/(?:[^/]+)'
SOCKET_PATH_REGEX = r'^/socket/(?P<viewer_token>[^/]+)'

class ClientCredentialsHandler(object):
    def __init__(self, io_loop, private_state, config_state, credentials_manager):
        self.private_state = private_state
        self.config_state = config_state
        self._on_changed_callback = (lambda: io_loop.add_callback(self._on_changed))
        private_state.add_changed_callback(self._on_changed_callback)
        self._previous_invalid_credentials = dict()
        self.credentials_manager = credentials_manager
        self._closed = False
        self.io_loop = io_loop

    def close(self):
        self.private_state.remove_changed_callback(self._on_changed_callback)
        self._closed = True

    def _on_changed(self):
        credentials = self.private_state.state.credentials
        try:
            for key, value in credentials.iteritems():
                prev_value = self._previous_invalid_credentials.get(key, 'invalid')
                if prev_value != value:
                    parsed_key = json.loads(key)
                    provider = self.credentials_manager.get(parsed_key['key'], parsed_key.get('parameters'))
                    if provider is not None:
                        def handle_credentials(f, key=key):
                            if self._closed:
                                return
                            try:
                                credentials = f.result()
                                def func(s):
                                    s.credentials[key] = credentials
                                self.config_state.retry_txn(func)
                            except:
                                import traceback
                                traceback.print_exc()

                        provider.get(value).add_done_callback(handle_credentials)
        except:
            import traceback
            traceback.print_exc()

class StateHandler(object):
    def __init__(self, state, io_loop, send_update, receive_updates=True):
        self.state = state
        self._send_update = send_update
        self._receive_updates = receive_updates
        self.io_loop = io_loop
        self._last_generation = None
        if send_update is not None:
            self._on_state_changed_callback = (
                lambda: self.io_loop.add_callback(self._on_state_changed))
            self.state.add_changed_callback(self._on_state_changed_callback)

    def _on_state_changed(self):
        """Invoked when the viewer state changes."""
        raw_state, generation = self.state.raw_state_and_generation
        if generation != self._last_generation:
            self._last_generation = generation
            self._send_update(raw_state, generation)

    def request_send_state(self, generation):
        if self._send_update is not None:
            self._last_generation = generation
            self._on_state_changed()

    def receive_update(self, raw_state, generation):
        if self._receive_updates:
            self._last_generation = generation
            self.state.set_state(raw_state, generation)

    def close(self):
        if self._send_update is not None:
            self.state.remove_changed_callback(self._on_state_changed_callback)
            del self._on_state_changed_callback


class SockJSHandler(sockjs.tornado.SockJSConnection):

    @property
    def io_loop(self):
        return self.session.server.io_loop

    def on_open(self, info):
        server = self.session.server.neuroglancer_server
        m = re.match(SOCKET_PATH_REGEX, info.path)
        if m is None:
            self.close()
            return

        viewer_token = self.viewer_token = m.group('viewer_token')
        viewer = self.viewer = server.viewers.get(viewer_token)

        if viewer is None:
            self.close()
            return

        private_state = self.private_state = trackable_state.TrackableState(
            viewer_config_state.PrivateState)

        managed_states = [
            dict(key='c', state=viewer.config_state, send_updates=True, receive_updates=False),
            dict(key='p', state=private_state, send_updates=False, receive_updates=True),
        ]
        if hasattr(viewer, 'shared_state'):
            managed_states.append(
                dict(key='s', state=viewer.shared_state, send_updates=True, receive_updates=True))

        self._state_handlers = dict()
        from .default_credentials_manager import default_credentials_manager
        self._credentials_handler = ClientCredentialsHandler(io_loop=self.io_loop,
                                                             private_state=private_state,
                                                             config_state=viewer.config_state,
                                                             credentials_manager=default_credentials_manager)
        self._next_action_id = 0

        def make_state_handler(key, state, send_updates, receive_updates):
            def send_update(raw_state, generation):
                if not self.is_open:
                    return
                message = {'t': 'setState', 'k': key, 's': raw_state, 'g': generation}
                self.send(encode_json(message))

            handler = StateHandler(
                state=state,
                io_loop=self.io_loop,
                send_update=send_update if send_updates else None,
                receive_updates=receive_updates)
            self._state_handlers[key] = handler

        for x in managed_states:
            make_state_handler(**x)

        self.is_open = True

    def on_message(self, message_text):
        if self.viewer is None:
            return
        try:
            message = decode_json(message_text)
            if isinstance(message, dict):
                t = message['t']
                if t == 'getState':
                    handler = self._state_handlers[message['k']]
                    handler.request_send_state(six.text_type(message['g']))
                    return
                if t == 'setState':
                    handler = self._state_handlers[message['k']]
                    handler.receive_update(message['s'], six.text_type(message['g']))
                    return
                if t == 'action':
                    new_next_action_id = int(message['id']) + 1
                    next_action_id = self._next_action_id
                    self._next_action_id = new_next_action_id
                    max_actions = new_next_action_id - next_action_id
                    actions = message['actions']
                    if max_actions == 0:
                        actions = []
                    else:
                        actions = actions[-max_actions:]
                    self.io_loop.add_callback(self.send, json.dumps({'t': 'ackAction', 'id': message['id']}))
                    for i, action in enumerate(actions):
                        self.io_loop.add_callback(self.viewer.actions.invoke, action['action'], action['state'])
        except:
            # import pdb
            # pdb.post_mortem()
            import traceback
            traceback.print_exc()
            # Ignore malformed JSON

    def on_close(self):
        viewer = self.viewer
        self.is_open = False
        if viewer is not None:
            for state_handler in six.itervalues(self._state_handlers):
                state_handler.close()
            self._credentials_handler.close()
        del self._state_handlers



# coding=utf-8
# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for controlling a browser that runs Neuroglancer."""

from __future__ import absolute_import

import time

class Webdriver(object):

    def __init__(self, viewer=None, headless=True, window_size=(1920, 1080)):
        import selenium.webdriver
        import selenium.webdriver.chrome.options
        import selenium.webdriver.common.desired_capabilities
        try:
            # Use chromedriver_binary package if available
            import chromedriver_binary
        except ImportError:
            # Fallback to system chromedriver
            pass
        if viewer is None:
            from .viewer import Viewer
            viewer = Viewer()
        self.viewer = viewer
        chrome_options = selenium.webdriver.chrome.options.Options()
        if headless:
            chrome_options.add_argument('--headless')
        chrome_options.add_argument('--window_size=%dx%d' % (window_size[0], window_size[1]))
        caps = selenium.webdriver.common.desired_capabilities.DesiredCapabilities.CHROME.copy()
        caps['goog:loggingPrefs'] = {'browser': 'ALL'}
        self.driver = selenium.webdriver.Chrome(options=chrome_options, desired_capabilities=caps)
        self.driver.get(viewer.get_viewer_url())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.driver.close()

    def get_log(self):
        return self.driver.get_log('browser')

    def sync(self):
        """Wait until client is ready."""
        while True:
            new_state = self.viewer.screenshot().viewer_state
            # Ensure self.viewer.state has also been updated to the new state.
            # The state sent in the screenshot reply can be newer.
            if new_state == self.viewer.state:
                return new_state
            time.sleep(0.1)

    @property
    def root_element(self):
        return self.driver.find_element_by_xpath('//body')

    def action_chain(self):
        import selenium.webdriver
        return selenium.webdriver.common.action_chains.ActionChains(self.driver)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from . import server, url_state, viewer_base


class _ViewerHelper(object):
    """Mixin for implementing viewers based on the built-in server."""

    def __init__(self):
        super(_ViewerHelper, self).__init__()
        server.register_viewer(self)

    def defer_callback(self, callback, *args, **kwargs):
        server.defer_callback(callback, *args, **kwargs)

    def __repr__(self):
        return self.get_viewer_url()

    def _repr_html_(self):
        return '<a href="%s" target="_blank">Viewer</a>' % self.get_viewer_url()

    def get_viewer_url(self):
        raise NotImplementedError


class Viewer(viewer_base.ViewerBase, _ViewerHelper):
    def __init__(self):
        super(Viewer, self).__init__()
        server.register_viewer(self)

    def get_viewer_url(self):
        return '%s/v/%s/' % (server.get_server_url(), self.token)


class UnsynchronizedViewer(viewer_base.UnsynchronizedViewerBase, _ViewerHelper):
    def __init__(self):
        super(UnsynchronizedViewer, self).__init__()
        server.register_viewer(self)

    def get_viewer_url(self):
        return url_state.to_url(self.raw_state, '%s/v/%s/' % (server.get_server_url(), self.token))



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

cube_corner_position_offsets = [
    [0, 0, 0],  #
    [1, 0, 0],
    [1, 1, 0],
    [0, 1, 0],
    [0, 0, 1],
    [1, 0, 1],
    [1, 1, 1],
    [0, 1, 1]
]

cube_edge_index_to_corner_index_pair_table = [
    [0, 1],
    [1, 2],
    [3, 2],
    [0, 3],
    [4, 5],
    [5, 6],
    [7, 6],
    [4, 7],
    [0, 4],
    [1, 5],
    [2, 6],
    [3, 7],
]

for edge_i, corners in enumerate(cube_edge_index_to_corner_index_pair_table):
    if cube_corner_position_offsets[corners[0]] > cube_corner_position_offsets[corners[1]]:
        print('edge %d is flipped' % (edge_i))

cube_edge_vertex_map_selectors_table = [0] * 256
cube_edge_mask_table = [0] * 256

for corners_present in range(256):
    selectors = 0
    edge_mask = 0
    for edge_i, corners in enumerate(cube_edge_index_to_corner_index_pair_table):
        edge_corners_present = [(corners_present >> corner_i) & 1 for corner_i in corners]
        if 0 in edge_corners_present and 1 in edge_corners_present:
            edge_mask |= (1 << edge_i)
        selector = edge_corners_present[0]
        selectors |= selector << edge_i
    cube_edge_mask_table[corners_present] = edge_mask
    cube_edge_vertex_map_selectors_table[corners_present] = selectors

print('static uint32_t cube_edge_mask_table[256] = {')
print(', '.join(map(hex, cube_edge_mask_table)))
print('};')

print('static uint32_t cube_edge_vertex_map_selectors_table[256] = {')
print(', '.join(map(hex, cube_edge_vertex_map_selectors_table)))
print('};')



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import concurrent.futures
import logging
import threading

from . import credentials_provider
from .futures import run_on_new_thread


class GoogleCredentialsProvider(credentials_provider.CredentialsProvider):
    def __init__(self, scopes, client_id, client_secret):
        super(GoogleCredentialsProvider, self).__init__()
        self.scopes = scopes
        self.client_id = client_id
        self.client_secret = client_secret

        # Make sure logging is initialized.  Does nothing if logging has already
        # been initialized.
        logging.basicConfig()

    def get_new(self):
        def func():
            import apitools.base.py.credentials_lib
            result = apitools.base.py.credentials_lib.GetCredentials(
                package_name='',
                scopes=self.scopes,
                client_id=self.client_id,
                client_secret=self.client_secret,
                user_agent=u'python-neuroglancer',
            )
            return dict(tokenType=u'Bearer', accessToken=result.get_access_token().access_token)

        return run_on_new_thread(func)



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import collections
import math
import threading

import numpy as np
import six

from . import downsample, downsample_scales
from .chunks import encode_jpeg, encode_npz, encode_raw
from . import trackable_state
from .random_token import make_random_token


class MeshImplementationNotAvailable(Exception):
    pass

class MeshesNotSupportedForVolume(Exception):
    pass

class InvalidObjectIdForMesh(Exception):
    pass

class LocalVolume(trackable_state.ChangeNotifier):
    def __init__(self,
                 data,
                 dimensions,
                 volume_type=None,
                 voxel_offset=None,
                 encoding='npz',
                 max_voxels_per_chunk_log2=None,
                 mesh_options=None,
                 downsampling='3d',
                 chunk_layout=None,
                 max_downsampling=downsample_scales.DEFAULT_MAX_DOWNSAMPLING,
                 max_downsampled_size=downsample_scales.DEFAULT_MAX_DOWNSAMPLED_SIZE,
                 max_downsampling_scales=downsample_scales.DEFAULT_MAX_DOWNSAMPLING_SCALES):
        """Initializes a LocalVolume.

        @param data: Source data.

        @param downsampling: '3d' to use isotropic downsampling, '2d' to downsample separately in
            XY, XZ, and YZ, None to use no downsampling.

        @param max_downsampling: Maximum amount by which on-the-fly downsampling may reduce the
            volume of a chunk.  For example, 4x4x4 downsampling reduces the volume by 64.

        @param volume_type: either 'image' or 'segmentation'.  If not specified, guessed from the
            data type.

        @param voxel_size: Sequence [x, y, z] of floats.  Specifies the voxel size.

        @param mesh_options: A dict with the following keys specifying options for mesh
            simplification for 'segmentation' volumes:

                - max_quadrics_error: float.  Edge collapses with a larger associated quadrics error
                  than this amount are prohibited.  Set this to a negative number to disable mesh
                  simplification, and just use the original mesh produced by the marching cubes
                  algorithm.  Defaults to 1e6.  The effect of this value depends on the voxel_size.

                - max_normal_angle_deviation: float.  Edge collapses that change a triangle normal
                  by more than this angle are prohibited.  The angle is specified in degrees.
                  Defaults to 90.

                - lock_boundary_vertices: bool.  Retain all vertices along mesh surface boundaries,
                  which can only occur at the boundary of the volume.  Defaults to true.
        """
        super(LocalVolume, self).__init__()
        self.token = make_random_token()
        self.data = data
        self.shape = data.shape
        rank = self.rank = len(self.shape)
        if rank != dimensions.rank:
            raise ValueError('rank of data (%d) must match rank of coordinate space (%d)' %
                             (rank, dimensions.rank))
        if voxel_offset is None:
            voxel_offset = np.zeros(rank, dtype=np.int64)
        else:
            voxel_offset = np.array(voxel_offset, dtype=np.int64)
        if voxel_offset.shape != (rank,):
            raise ValueError('voxel_offset must have shape of (%d,)' % (rank,))
        self.voxel_offset = voxel_offset
        self.dimensions = dimensions
        self.data_type = np.dtype(data.dtype).name
        if self.data_type == 'float64':
            self.data_type = 'float32'
        self.encoding = encoding
        if volume_type is None:
            if self.rank == 3 and (self.data_type == 'uint16' or
                                   self.data_type == 'uint32' or
                                   self.data_type == 'uint64'):
                volume_type = 'segmentation'
            else:
                volume_type = 'image'
        self.volume_type = volume_type

        self._mesh_generator = None
        self._mesh_generator_pending = None
        self._mesh_generator_lock = threading.Condition()
        self._mesh_options = mesh_options.copy() if mesh_options is not None else dict()

        self.max_voxels_per_chunk_log2 = max_voxels_per_chunk_log2

        self.downsampling_layout = downsampling
        if chunk_layout is None:
            if downsampling == '2d':
                chunk_layout = 'flat'
            else:
                chunk_layout = 'isotropic'
        self.chunk_layout = chunk_layout

        self.max_downsampling = max_downsampling
        self.max_downsampled_size = max_downsampled_size
        self.max_downsampling_scales = max_downsampling_scales

    def info(self):
        info = dict(dataType=self.data_type,
                    encoding=self.encoding,
                    generation=self.change_count,
                    coordinateSpace=self.dimensions.to_json(),
                    shape=self.shape,
                    volumeType=self.volume_type,
                    voxelOffset=self.voxel_offset,
                    chunkLayout=self.chunk_layout,
                    downsamplingLayout=self.downsampling_layout,
                    maxDownsampling=None if math.isinf(self.max_downsampling) else self.max_downsampling,
                    maxDownsampledSize=None if math.isinf(self.max_downsampled_size) else self.max_downsampled_size,
                    maxDownsamplingScales=None if math.isinf(self.max_downsampling_scales) else self.max_downsampling_scales,
        )
        if self.max_voxels_per_chunk_log2 is not None:
            info['maxVoxelsPerChunkLog2'] = self.max_voxels_per_chunk_log2

        return info

    def get_encoded_subvolume(self, data_format, start, end, scale_key):
        rank = self.rank
        if len(start) != rank or len(end) != rank:
            raise ValueError('Invalid request')
        downsample_factor = np.array(scale_key.split(','), dtype=np.int64)
        if (len(downsample_factor) != rank or np.any(downsample_factor < 1)
            or np.any(downsample_factor > self.max_downsampling)
            or np.prod(downsample_factor) > self.max_downsampling):
            raise ValueError('Invalid downsampling factor.')
        downsampled_shape = np.cast[np.int64](np.ceil(self.shape / downsample_factor))
        if np.any(end < start) or np.any(start < 0) or np.any(end > downsampled_shape):
            raise ValueError('Out of bounds data request.')

        indexing_expr = tuple(np.s_[start[i] * downsample_factor[i]:end[i] * downsample_factor[i]]
                              for i in range(rank))
        subvol = self.data[indexing_expr]
        if subvol.dtype == 'float64':
            subvol = np.cast[np.float32](subvol)

        if np.any(downsample_factor != 1):
            if self.volume_type == 'image':
                subvol = downsample.downsample_with_averaging(subvol, downsample_factor)
            else:
                subvol = downsample.downsample_with_striding(subvol, downsample_factor)
        content_type = 'application/octet-stream'
        if data_format == 'jpeg':
            data = encode_jpeg(subvol)
            content_type = 'image/jpeg'
        elif data_format == 'npz':
            data = encode_npz(subvol)
        elif data_format == 'raw':
            data = encode_raw(subvol)
        else:
            raise ValueError('Invalid data format requested.')
        return data, content_type

    def get_object_mesh(self, object_id):
        mesh_generator = self._get_mesh_generator()
        data = mesh_generator.get_mesh(object_id)
        if data is None:
            raise InvalidObjectIdForMesh()
        return data

    def _get_mesh_generator(self):
        if self._mesh_generator is not None:
            return self._mesh_generator
        while True:
            with self._mesh_generator_lock:
                if self._mesh_generator is not None:
                    return self._mesh_generator
                if self._mesh_generator_pending is not None:
                    while self._mesh_generator is None:
                        self._mesh_generator_lock.wait()
                    if self._mesh_generator is not None:
                        return self._mesh_generator
                try:
                    from . import _neuroglancer
                except ImportError:
                    raise MeshImplementationNotAvailable()
                if not (self.rank == 3 and
                        (self.data_type == 'uint8' or self.data_type == 'uint16' or
                         self.data_type == 'uint32' or self.data_type == 'uint64')):
                    raise MeshesNotSupportedForVolume()
                pending_obj = object()
                self._mesh_generator_pending = pending_obj
            data = self.data
            new_mesh_generator = _neuroglancer.OnDemandObjectMeshGenerator(
                data.transpose(),
                self.dimensions.scales, np.zeros(3), **self._mesh_options)
            with self._mesh_generator_lock:
                if self._mesh_generator_pending is not pending_obj:
                    continue
                self._mesh_generator = new_mesh_generator
                self._mesh_generator_pending = False
                self._mesh_generator_lock.notify_all()
            return new_mesh_generator

    def __deepcopy__(self, memo):
        """Since this type is immutable, we don't need to deepcopy it.

        Actually deep copying would intefere with the use of deepcopy by JsonObjectWrapper.
        """
        return self

    def invalidate(self):
        """Mark the data invalidated.  Clients will refetch the volume."""
        with self._mesh_generator_lock:
            self._mesh_generator_pending = None
            self._mesh_generator = None
        self._dispatch_changed_callbacks()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import concurrent.futures
import json
import multiprocessing
import re
import socket
import threading
import weakref

import numpy as np
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web

import sockjs.tornado


from . import local_volume, static
from . import skeleton
from .json_utils import json_encoder_default
from .random_token import make_random_token
from .sockjs_handler import SOCKET_PATH_REGEX, SOCKET_PATH_REGEX_WITHOUT_GROUP, SockJSHandler

INFO_PATH_REGEX = r'^/neuroglancer/info/(?P<token>[^/]+)$'
SKELETON_INFO_PATH_REGEX = r'^/neuroglancer/skeletoninfo/(?P<token>[^/]+)$'

DATA_PATH_REGEX = r'^/neuroglancer/(?P<data_format>[^/]+)/(?P<token>[^/]+)/(?P<scale_key>[^/]+)/(?P<start>[0-9]+(?:,[0-9]+)*)/(?P<end>[0-9]+(?:,[0-9]+)*)$'

SKELETON_PATH_REGEX = r'^/neuroglancer/skeleton/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

MESH_PATH_REGEX = r'^/neuroglancer/mesh/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

STATIC_PATH_REGEX = r'^/v/(?P<viewer_token>[^/]+)/(?P<path>(?:[a-zA-Z0-9_\-][a-zA-Z0-9_\-.]*)?)$'

global_static_content_source = None

global_server_args = dict(bind_address='127.0.0.1', bind_port=0)

debug = False

class Server(object):
    def __init__(self, ioloop, bind_address='127.0.0.1', bind_port=0):
        self.viewers = weakref.WeakValueDictionary()
        self.token = make_random_token()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=multiprocessing.cpu_count())

        self.ioloop = ioloop
        sockjs_router = sockjs.tornado.SockJSRouter(
            SockJSHandler, SOCKET_PATH_REGEX_WITHOUT_GROUP, io_loop=ioloop)
        sockjs_router.neuroglancer_server = self
        def log_function(handler):
            if debug:
                print("%d %s %.2fs" % (handler.get_status(),
                                       handler.request.uri, handler.request.request_time()))

        app = self.app = tornado.web.Application(
            [
                (STATIC_PATH_REGEX, StaticPathHandler, dict(server=self)),
                (INFO_PATH_REGEX, VolumeInfoHandler, dict(server=self)),
                (SKELETON_INFO_PATH_REGEX, SkeletonInfoHandler, dict(server=self)),
                (DATA_PATH_REGEX, SubvolumeHandler, dict(server=self)),
                (SKELETON_PATH_REGEX, SkeletonHandler, dict(server=self)),
                (MESH_PATH_REGEX, MeshHandler, dict(server=self)),
            ] + sockjs_router.urls,
            log_function=log_function,
            # Set a large maximum message size to accommodate large screenshot
            # messages.
            websocket_max_message_size=100 * 1024 * 1024)
        http_server = tornado.httpserver.HTTPServer(app)
        sockets = tornado.netutil.bind_sockets(port=bind_port, address=bind_address)
        http_server.add_sockets(sockets)
        actual_port = sockets[0].getsockname()[1]

        global global_static_content_source
        if global_static_content_source is None:
            global_static_content_source = static.get_default_static_content_source()

        if bind_address == '0.0.0.0' or bind_address == '::':
            hostname = socket.getfqdn()
        else:
            hostname = bind_address

        self.server_url = 'http://%s:%s' % (hostname, actual_port)

    def get_volume(self, key):
        dot_index = key.find('.')
        if dot_index == -1:
            return None
        viewer_token = key[:dot_index]
        volume_token = key[dot_index+1:]
        viewer = self.viewers.get(viewer_token)
        if viewer is None:
            return None
        return viewer.volume_manager.volumes.get(volume_token)


class BaseRequestHandler(tornado.web.RequestHandler):
    def initialize(self, server):
        self.server = server

class StaticPathHandler(BaseRequestHandler):
    def get(self, viewer_token, path):
        if viewer_token != self.server.token and viewer_token not in self.server.viewers:
            self.send_error(404)
            return
        try:
            data, content_type = global_static_content_source.get(path)
        except ValueError as e:
            self.send_error(404, message=e.args[0])
            return
        self.set_header('Content-type', content_type)
        self.finish(data)


class VolumeInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())

class SkeletonInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())

class SubvolumeHandler(BaseRequestHandler):
    @tornado.web.asynchronous
    def get(self, data_format, token, scale_key, start, end):
        start_pos = np.array(start.split(','), dtype=np.int64)
        end_pos = np.array(end.split(','), dtype=np.int64)
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        def handle_subvolume_result(f):
            try:
                data, content_type = f.result()
            except ValueError as e:
                self.send_error(400, message=e.args[0])
                return

            self.set_header('Content-type', content_type)
            self.finish(data)

        self.server.executor.submit(
            vol.get_encoded_subvolume,
            data_format=data_format, start=start_pos, end=end_pos, scale_key=scale_key).add_done_callback(
                lambda f: self.server.ioloop.add_callback(lambda: handle_subvolume_result(f)))


class MeshHandler(BaseRequestHandler):
    @tornado.web.asynchronous
    def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        def handle_mesh_result(f):
            try:
                encoded_mesh = f.result()
            except local_volume.MeshImplementationNotAvailable:
                self.send_error(501, message='Mesh implementation not available')
                return
            except local_volume.MeshesNotSupportedForVolume:
                self.send_error(405, message='Meshes not supported for volume')
                return
            except local_volume.InvalidObjectIdForMesh:
                self.send_error(404, message='Mesh not available for specified object id')
                return
            except ValueError as e:
                self.send_error(400, message=e.args[0])
                return

            self.set_header('Content-type', 'application/octet-stream')
            self.finish(encoded_mesh)

        self.server.executor.submit(vol.get_object_mesh, object_id).add_done_callback(
            lambda f: self.server.ioloop.add_callback(lambda: handle_mesh_result(f)))


class SkeletonHandler(BaseRequestHandler):
    @tornado.web.asynchronous
    def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)

        def handle_result(f):
            try:
                encoded_skeleton = f.result()
            except:
                self.send_error(500, message=e.args[0])
                return
            if encoded_skeleton is None:
                self.send_error(404, message='Skeleton not available for specified object id')
                return
            self.set_header('Content-type', 'application/octet-stream')
            self.finish(encoded_skeleton)

        def get_encoded_skeleton(skeletons, object_id):
            skeleton = skeletons.get_skeleton(object_id)
            if skeleton is None:
                return None
            return skeleton.encode(skeletons)

        self.server.executor.submit(
            get_encoded_skeleton, vol, object_id).add_done_callback(
                lambda f: self.server.ioloop.add_callback(lambda: handle_result(f)))


global_server = None


def set_static_content_source(*args, **kwargs):
    global global_static_content_source
    global_static_content_source = static.get_static_content_source(*args, **kwargs)


def set_server_bind_address(bind_address='127.0.0.1', bind_port=0):
    global global_server_args
    global_server_args = dict(bind_address=bind_address, bind_port=bind_port)


def is_server_running():
    return global_server is not None


def stop():
    """Stop the server, invalidating any viewer URLs.

    This allows any previously-referenced data arrays to be garbage collected if there are no other
    references to them.
    """
    global global_server
    if global_server is not None:
        ioloop = global_server.ioloop
        def stop_ioloop():
            ioloop.stop()
            ioloop.close()
        global_server.ioloop.add_callback(stop_ioloop)
        global_server = None


def get_server_url():
    return global_server.server_url


_global_server_lock = threading.Lock()

def start():
    global global_server
    with _global_server_lock:
        if global_server is not None: return
        done = threading.Event()
        def start_server():
            global global_server
            ioloop = tornado.ioloop.IOLoop()
            ioloop.make_current()
            global_server = Server(ioloop=ioloop, **global_server_args)
            done.set()
            ioloop.start()
        thread = threading.Thread(target=start_server)
        thread.daemon = True
        thread.start()
        done.wait()

def register_viewer(viewer):
    start()
    global_server.viewers[viewer.token] = viewer

def defer_callback(callback, *args, **kwargs):
    """Register `callback` to run in the server event loop thread."""
    start()
    global_server.ioloop.add_callback(lambda: callback(*args, **kwargs))



#!/usr/bin/env python
from __future__ import print_function

import os
import shutil
import subprocess
import platform
from distutils.command.build import build
from setuptools.command.build_ext import build_ext

from setuptools import Extension, find_packages, setup

class build_ext_subclass(build_ext):
    def finalize_options(self):
        build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


static_files = [
    'main.bundle.js',
    'main.css',
    'chunk_worker.bundle.js',
    'async_computation.bundle.js',
    'draco.bundle.js',
    'index.html',
]


class bundle_client(build):

    user_options = [
        ('client-bundle-type=', None,
         'The nodejs bundle type. "min" (default) creates condensed static files for production, "dev" creates human-readable files.'
         )
    ]

    def initialize_options(self):

        self.client_bundle_type = 'min'

    def finalize_options(self):

        if self.client_bundle_type not in ['min', 'dev']:
            raise RuntimeError('client-bundle-type has to be one of "min" or "dev"')

    def run(self):

        this_dir = os.path.abspath(os.path.dirname(__file__))
        project_dir = os.path.join(this_dir, '..')

        build_dir = os.path.join(project_dir, 'dist/python-' + self.client_bundle_type)
        static_dir = os.path.join(this_dir, 'neuroglancer/static')

        print("Project dir " + project_dir)
        print("Build dir " + build_dir)
        print("Static dir " + static_dir)

        prev_dir = os.path.abspath('.')
        os.chdir(project_dir)

        target = {"min": "build-python-min", "dev": "build-python"}

        try:
            t = target[self.client_bundle_type]
            subprocess.call('npm i', shell=True)
            res = subprocess.call('npm run %s' % t, shell=True)
        except:
            raise RuntimeError(
                'Could not run \'npm run %s\'. Make sure node.js >= v5.9.0 is installed and in your path.'
                % t)

        if res != 0:
            raise RuntimeError('failed to bundle neuroglancer node.js project')

        try:
            os.mkdir(static_dir)
        except OSError:
            pass

        for f in static_files:
            shutil.copy(os.path.join(build_dir, f), os.path.join(static_dir, f))

        os.chdir(prev_dir)

setup_dir = os.path.dirname(__file__)
src_dir = os.path.join(setup_dir, 'ext/src')
openmesh_dir = os.path.join(setup_dir, 'ext/third_party/openmesh/OpenMesh/src')
local_sources = [
    '_neuroglancer.cc',
    'openmesh_dependencies.cc',
    'on_demand_object_mesh_generator.cc',
    'voxel_mesh_generator.cc',
    'mesh_objects.cc',
]

USE_OMP = False
if USE_OMP:
    openmp_flags = ['-fopenmp']
else:
    openmp_flags = []

extra_compile_args = ['-std=c++11', '-fvisibility=hidden', '-O3'] + openmp_flags
if platform.system() == 'Darwin':
    extra_compile_args.insert(0, '-stdlib=libc++')

setup(
    name='neuroglancer',
    version='2.4',
    description='Python data backend for neuroglancer, a WebGL-based viewer for volumetric data',
    author='Jeremy Maitin-Shepard, Jan Funke',
    author_email='jbms@google.com, jfunke@iri.upc.edu',
    url='https://github.com/google/neuroglancer',
    license='Apache License 2.0',
    packages=find_packages(),
    package_data={
        'neuroglancer.static': static_files,
    },
    setup_requires=[
        "numpy>=1.11.0",
    ],
    install_requires=[
        "Pillow>=3.2.0",
        "numpy>=1.11.0",
        'requests',
        'tornado<6',
        'sockjs-tornado',
        'six',
        'google-apitools',
    ],
    test_requires=[
        'pytest',
        'selenium',
        'chromedriver-binary',
    ],
    extras_require={
        ":python_version<'3.2'": ['futures'],
    },
    ext_modules=[
        Extension(
            'neuroglancer._neuroglancer',
            sources=[os.path.join(src_dir, name) for name in local_sources],
            language='c++',
            include_dirs=[openmesh_dir],
            define_macros=[
                ('_USE_MATH_DEFINES', None),  # Needed by OpenMesh when used with MSVC
            ],
            extra_compile_args=extra_compile_args,
            extra_link_args=openmp_flags),
    ],
    cmdclass={
        'bundle_client': bundle_client,
        'build_ext': build_ext_subclass,
    },
)



#!/usr/bin/env python

from __future__ import print_function

import argparse
import collections
import uuid
import copy
import heapq
import json
import re
import sqlite3
import logging
import os

import numpy as np
import six

import neuroglancer

debug_graph = False
verbose_merging = False


def normalize_edge(e):
    id_a, id_b = e
    if id_a > id_b:
        id_a, id_b = id_b, id_a
    return id_a, id_b


class GreedyMulticut(object):
    def __init__(self, combine_edges, edge_priority):
        # Contains (score, edge_map_value) tuple values in heap order.  The
        # edge_map_value is the actual corresponding value in edge_map, not a copy.
        self.edge_heap = []

        # Maps segment_id -> set of segment_id neighbors.
        self.regions = dict()

        # Maps (id_a, id_b) -> edge_map_value=[score, key, edge_object]
        self.edge_map = dict()
        self.combine_edges = combine_edges
        self.edge_priority = edge_priority
        self.num_valid_edges = 0
        self._initialized = False

    def add_edge(self, e, edge):
        id_a, id_b = e
        id_a, id_b = normalize_edge((id_a, id_b))
        self.regions.setdefault(id_a, set()).add(id_b)
        self.regions.setdefault(id_b, set()).add(id_a)
        key = (id_a, id_b)
        entry = self.edge_map.get(key, None)
        if entry is not None:
            edge_data = entry[2] = self.combine_edges(entry[0], edge)
            entry[0] = self.edge_priority(edge_data)
        else:
            entry = self.edge_map[key] = [self.edge_priority(edge), key, edge]
            self.num_valid_edges += 1
        if self._initialized:
            self._add_to_heap(entry)

    def _initialize_heap(self):
        if self._initialized:
            return
        for key in self.edge_map:
            entry = self.edge_map[key]
            self._add_to_heap(entry)
        self._initialized = True

    def _add_to_heap(self, entry):
        heapq.heappush(self.edge_heap, (entry[0], entry))

    def remove_edge_from_heap(self, segment_ids):
        """Remove an edge from the heap."""
        self._initialize_heap()
        key = normalize_edge(segment_ids)
        if key in self.edge_map:
            self.edge_map[key][0] = None
            self.num_valid_edges -= 1

    def check_consistency(self):
        self._initialize_heap()
        expected_regions = dict()
        for key, entry in six.viewitems(self.edge_map):
            assert entry[1] == key
            expected_regions.setdefault(key[0], set()).add(key[1])
            expected_regions.setdefault(key[1], set()).add(key[0])

        assert expected_regions == self.regions

        num_valid_edges = 0
        for e in self.edge_heap:
            if self._is_valid_heap_entry(e):
                num_valid_edges += 1
        assert num_valid_edges == self.num_valid_edges

    def merge(self, e):
        id_a, id_b = e
        self._initialize_heap()
        id_a, id_b = normalize_edge((id_a, id_b))
        if (id_a, id_b) not in self.edge_map:
            raise KeyError
        for neighbor in self.regions[id_b]:
            if neighbor == id_a:
                continue
            expired_ids = normalize_edge((neighbor, id_b))
            new_ids = normalize_edge((neighbor, id_a))
            new_edge = self.edge_map.get(new_ids)
            expired_edge = self.edge_map[expired_ids]
            if new_edge is not None:
                edge_data = new_edge[2] = self.combine_edges(new_edge[2], expired_edge[2])
                if new_edge[0] is not None:
                    self.num_valid_edges -= 1
                if expired_edge[0] is not None:
                    self.num_valid_edges -= 1
                self.num_valid_edges += 1
                new_edge[0] = self.edge_priority(edge_data)
                self._add_to_heap(new_edge)
            else:
                self.regions[neighbor].add(id_a)
                self.regions[id_a].add(neighbor)
                self.edge_map[new_ids] = expired_edge
                expired_edge[1] = new_ids
                # No need to add to heap, since score hasn't changed.
            del self.edge_map[expired_ids]
            self.regions[neighbor].remove(id_b)
        del self.regions[id_b]
        self.regions[id_a].remove(id_b)
        del self.edge_map[(id_a, id_b)]
        self.num_valid_edges -= 1

    def _is_valid_heap_entry(self, heap_entry):
        score, entry = heap_entry
        expected_entry = self.edge_map.get(entry[1])
        if entry is not expected_entry or entry[0] is not score:
            return None
        return entry

    def get_next_edge(self):
        self._initialize_heap()
        while True:
            if self.num_valid_edges == 0:
                return None
            heap_entry = self.edge_heap[0]
            entry = self._is_valid_heap_entry(heap_entry)
            if entry is None:
                heapq.heappop(self.edge_heap)
            else:
                return entry


Edge = collections.namedtuple('Edge', ['segment_ids', 'score', 'position'])


def load_edges(path):
    edges = []
    with open(path, 'r') as f:
        f.readline()
        for line in f:
            parts = line.split(',')
            segment_a = int(parts[0].strip())
            segment_b = int(parts[1].strip())
            score = float(parts[2].strip())
            position = (int(parts[3].strip()), int(parts[4].strip()), int(parts[5].strip()))
            edges.append(Edge(segment_ids=(segment_a, segment_b), score=score, position=position))
    return edges


def load_split_seeds(path):
    with open(path, 'r') as f:
        raw_seeds = json.loads(f.read())
    seeds = collections.OrderedDict()
    for component in raw_seeds:
        seeds.setdefault(component['label'], []).extend(component['supervoxels'])
    return seeds


def build_graph(edges):
    logging.info('Building graph with %d edges', len(edges))

    def combine_edges(a, b):
        return a + b

    def edge_priority(x):
        return x

    greedy_multicut = GreedyMulticut(
        combine_edges=combine_edges,
        edge_priority=edge_priority,
    )
    for edge in edges:
        greedy_multicut.add_edge(edge.segment_ids, edge.score)
    return greedy_multicut


class AgglomerationGraph(object):
    def __init__(self, conn):
        self.conn = conn
        self.agglo_members_cache = dict()
        self.agglo_edges_cache = dict()

    def get_agglo_id(self, supervoxel_id):
        c = self.conn.cursor()
        c.execute('SELECT agglo_id FROM supervoxels WHERE supervoxel_id=?', (int(supervoxel_id), ))
        result = c.fetchone()
        if result is None:
            return supervoxel_id
        return result[0]

    def get_agglo_members(self, agglo_id):
        result = self.agglo_members_cache.get(agglo_id)
        if result is not None:
            return result
        c = self.conn.cursor()
        c.execute('SELECT supervoxel_id FROM supervoxels WHERE agglo_id=?', (int(agglo_id), ))
        result = [row[0] for row in c.fetchall()]
        self.agglo_members_cache[agglo_id] = result
        return result

    def get_agglo_edges(self, agglo_id):
        result = self.agglo_edges_cache.get(agglo_id)
        if result is not None:
            return result
        c = self.conn.cursor()
        c.execute('SELECT segment_a, segment_b, score, x, y, z FROM edges WHERE agglo_id=?',
                  (int(agglo_id), ))
        result = [
            Edge(segment_ids=(row[0], row[1]), score=row[2], position=(row[3], row[4], row[5]))
            for row in c.fetchall()
        ]
        self.agglo_edges_cache[agglo_id] = result
        return result


def _make_supervoxel_map(graph, split_seeds, need_agglo_ids):
    supervoxel_map = dict()
    agglo_ids = dict()

    for label in [0, 1]:
        for seed in split_seeds[label]:
            supervoxel_id = seed['supervoxel_id']
            if need_agglo_ids:
                agglo_id = graph.get_agglo_id(supervoxel_id)
                if agglo_id == 0:
                    continue
                agglo_ids.setdefault(agglo_id, []).append((label, seed))
            supervoxel_map.setdefault(supervoxel_id, set()).add(label)
    return agglo_ids, supervoxel_map


def do_split(graph, split_seeds, agglo_id=None, supervoxels=None):

    agglo_ids, supervoxel_map = _make_supervoxel_map(graph, split_seeds, need_agglo_ids=agglo_id is None)

    if agglo_id is None:

        agglo_id_counts = {
            agglo_id: sum(z[1]['count'] for z in seeds)
            for agglo_id, seeds in six.viewitems(agglo_ids)
        }

        agglo_id = max(agglo_ids, key=lambda x: agglo_id_counts[x])

        if len(agglo_ids) > 1:
            logging.info('Warning: more than one agglomerated component.  ' +
                         'Choosing component %d with maximum number of seed points.', agglo_id)
            logging.info('agglo_id_counts = %r', agglo_id_counts)

    input_edges = graph.get_agglo_edges(agglo_id)
    if supervoxels is not None:
        input_edges = [x for x in input_edges if x.segment_ids[0] in supervoxels and x.segment_ids[1] in supervoxels]
    graph = build_graph(input_edges)
    if debug_graph:
        graph.check_consistency()

    cur_eqs = neuroglancer.EquivalenceMap()
    logging.info('Agglomerating')
    threshold = float('inf')
    while True:
        entry = graph.get_next_edge()
        if entry is None:
            if verbose_merging:
                logging.info('Stopping because entry is None')
            break
        if entry[0] > threshold:
            if verbose_merging:
                logging.info('Stopping because edge score %r is > threshold %r', entry[0],
                             threshold)
            break
        segment_ids = entry[1]
        seeds_a = supervoxel_map.get(segment_ids[0])
        seeds_b = supervoxel_map.get(segment_ids[1])
        if ((seeds_a is not None and len(seeds_a) > 1) or (seeds_b is not None and len(seeds_b) > 1)
                or (seeds_a is not None and seeds_b is not None and seeds_a != seeds_b)):
            if verbose_merging:
                logging.info('Excluding edge %r because of seeds: %r %r', segment_ids, seeds_a,
                             seeds_b)
            graph.remove_edge_from_heap(segment_ids)
            continue
        if verbose_merging:
            logging.info('Merging %r with score %r', segment_ids, entry[0])
        graph.merge(segment_ids)
        if debug_graph:
            graph.check_consistency()

        new_id = cur_eqs.union(*segment_ids)
        new_seeds = seeds_a or seeds_b
        if new_seeds:
            supervoxel_map[new_id] = new_seeds

    return dict(agglo_id=agglo_id, cur_eqs=cur_eqs, supervoxel_map=supervoxel_map)


def display_split_result(graph, agglo_id, cur_eqs, supervoxel_map, split_seeds, image_url,
                         segmentation_url):

    agglo_members = set(graph.get_agglo_members(agglo_id))
    state = neuroglancer.ViewerState()
    state.layers.append(name='image', layer=neuroglancer.ImageLayer(source=image_url))
    state.layers.append(
        name='original',
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            segments=agglo_members,
        ),
        visible=False,
    )
    state.layers.append(
        name='isolated-supervoxels',
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            segments=set(x for x, seeds in six.viewitems(supervoxel_map) if len(seeds) > 1),
        ),
        visible=False,
    )
    state.layers.append(
        name='split',
        layer=neuroglancer.SegmentationLayer(
            source=segmentation_url,
            equivalences=cur_eqs,
            segments=set(cur_eqs[x] for x in agglo_members),
        ))
    for label, component in six.viewitems(split_seeds):
        state.layers.append(
            name='seed%d' % label,
            layer=neuroglancer.PointAnnotationLayer(
                points=[seed['position'] for seed in component],
            ),
        )

    state.show_slices = False
    state.layout = '3d'
    all_seed_points = [
        seed['position'] for component in six.viewvalues(split_seeds) for seed in component
    ]
    state.voxel_coordinates = np.mean(all_seed_points, axis=0)
    return state


def _set_viewer_seeds(s, seeds):
    for inclusive in [False, True]:
        layer_name = 'inclusive-seeds' if inclusive else 'exclusive-seeds'
        s.layers[layer_name] = neuroglancer.AnnotationLayer(
            annotation_color='green' if inclusive else 'red',
            annotations=[
                dict(
                    type='point',
                    id=x['id'],
                    point=x['position'],
                    description=str(x['supervoxel_id']),
                ) for x in seeds[inclusive]
            ],
        )


def _get_viewer_seeds(s):
    seeds = [[], []]
    for inclusive in [False, True]:
        layer_name = 'inclusive-seeds' if inclusive else 'exclusive-seeds'
        try:
            layer = s.layers[layer_name]
        except KeyError:
            pass
        for x in layer.annotations:
            seeds[inclusive].append(
                dict(
                    id=x.id,
                    supervoxel_id=int(x.description),
                    position=tuple(map(int, x.point)),
                ))
    return seeds


class ComponentState(object):
    def __init__(self, data=None):
        self.supervoxels = set()
        self.seeds = [[], []]
        if data is not None:
            self.load(data)

    def load(self, data):
        self.supervoxels = set(data['supervoxels'])
        self.seeds = data['seeds']

    def to_json(self):
        return {
            'supervoxels': sorted(self.supervoxels),
            'seeds': self.seeds,
        }


class InteractiveState(object):
    def __init__(self, path):
        self.unused_supervoxels = set()
        self.components = []
        self.path = path
        self.selected_component = None

    def load(self):
        with open(self.path, 'r') as f:
            data = json.load(f)
            self.unused_supervoxels = set(data['unused_supervoxels'])
            self.components = map(ComponentState, data['components'])
            self.selected_component = data['selected_component']

    def initialize(self, supervoxel_ids):
        self.unused_supervoxels = set(supervoxel_ids)
        self.components = []
        self.selected_component = None

    def to_json(self):
        return {
            'unused_supervoxels': sorted(self.unused_supervoxels),
            'components': [x.to_json() for x in self.components],
            'selected_component': self.selected_component,
        }

    def save(self):
        if self.path is None:
            return
        tmp_path = self.path + '.tmp'
        with open(tmp_path, 'w') as f:
            f.write(json.dumps(self.to_json()))
        os.rename(tmp_path, self.path)

    def make_new_component(self):
        c = ComponentState()
        c.supervoxels = self.unused_supervoxels
        self.unused_supervoxels = set()
        self.selected_component = len(self.components)
        self.components.append(c)

    def cycle_selected_component(self, amount):
        if len(self.components) == 0:
            return
        if self.selected_component is None:
            if amount > 0:
                self.selected_component = 0
            else:
                self.selected_component = len(self.components) - 1
        else:
            self.selected_component = (
                self.selected_component + amount + len(self.components)) % len(self.components)

    def add_seed(self, supervoxel_id, position, inclusive):
        if self.selected_component is None:
            return
        c = self.components[self.selected_component]
        c.seeds[inclusive].append(
            dict(
                supervoxel_id=supervoxel_id,
                position=position,
                id=uuid.uuid4().hex))

class CachedSplitResult(object):
    def __init__(self, state, graph, agglo_id):
        self.state = state
        self.graph = graph
        self.agglo_id = agglo_id
        self.reset()

    def reset(self):
        self.selected_component = None
        self.seeds = [[], []]
        self.supervoxels = set()
        self.split_result = None

    def update(self):
        selected_component = self.state.selected_component
        if selected_component is None:
            if self.selected_component is None:
                return False
            self.reset()
            return True
        component = self.state.components[selected_component]
        if selected_component == self.selected_component:
            if self.supervoxels == component.supervoxels:
                if self.seeds == component.seeds:
                    return False
        self.selected_component = self.state.selected_component
        self.seeds = copy.deepcopy(component.seeds)
        self.supervoxels = set(component.supervoxels)
        print('Recomputing split result')
        self.split_result = do_split(
            graph=self.graph, split_seeds=self.seeds, agglo_id=self.agglo_id,
            supervoxels=self.supervoxels)
        print('Done recomputing split result')
        return True


class InteractiveSplitter(object):
    def __init__(self, graph, agglo_id, image_url, segmentation_url, state_path):
        self.graph = graph
        self.agglo_id = agglo_id
        self.image_url = image_url
        self.segmentation_url = segmentation_url
        self.state = InteractiveState(state_path)
        self.cached_split_result = CachedSplitResult(
            state=self.state, graph=self.graph, agglo_id=self.agglo_id)
        self.agglo_members = set(self.graph.get_agglo_members(agglo_id))

        if state_path is not None and os.path.exists(state_path):
            self.state.load()
        else:
            self.state.initialize(self.agglo_members)

        viewer = self.viewer = neuroglancer.Viewer()
        viewer.actions.add('inclusive-seed', self._add_inclusive_seed)
        viewer.actions.add('exclusive-seed', self._add_exclusive_seed)
        viewer.actions.add('next-component', self._next_component)
        viewer.actions.add('prev-component', self._prev_component)
        viewer.actions.add('new-component', self._make_new_component)
        viewer.actions.add('exclude-component', self._exclude_component)
        viewer.actions.add('exclude-all-but-component', self._exclude_all_but_component)

        key_bindings = [
            ['bracketleft', 'prev-component'],
            ['bracketright', 'next-component'],
            ['at:dblclick0', 'exclude-component'],
            ['at:shift+mousedown2', 'exclude-all-but-component'],
            ['at:control+mousedown0', 'inclusive-seed'],
            ['at:shift+mousedown0', 'exclusive-seed'],
            ['enter', 'new-component'],
        ]

        with viewer.txn() as s:
            s.layers.append(
                name='image',
                layer=neuroglancer.ImageLayer(source=self.image_url),
            )
            s.layers.append(
                name='original',
                layer=neuroglancer.SegmentationLayer(
                    source=self.segmentation_url,
                    segments=self.agglo_members,
                ),
            )
            s.layers.append(
                name='unused',
                layer=neuroglancer.SegmentationLayer(source=self.segmentation_url,
                                                     ),
                visible=False,
            )
            s.layers.append(
                name='split-result',
                layer=neuroglancer.SegmentationLayer(
                    source=self.segmentation_url,
                    segments=self.agglo_members,
                ),
            )
            s.concurrent_downloads = 256
            self._update_state(s)

        with viewer.config_state.txn() as s:
            s.status_messages['help'] = ('KEYS: ' + ' | '.join('%s=%s' % (key, command)
                                                               for key, command in key_bindings))
            for key, command in key_bindings:
                s.input_event_bindings.viewer[key] = command
                s.input_event_bindings.slice_view[key] = command
                s.input_event_bindings.perspective_view[key] = command
            self._update_config_state(s)

        viewer.shared_state.add_changed_callback(
            lambda: viewer.defer_callback(self._handle_state_changed))

    def _add_inclusive_seed(self, s):
        self._add_seed(s, True)

    def _add_exclusive_seed(self, s):
        self._add_seed(s, False)

    def _exclude_component(self, s):
        if self.state.selected_component is None:
            return

        component = self.state.components[self.state.selected_component]
        supervoxel_id = self._get_mouse_supervoxel(s)

        if supervoxel_id is None:
            return

        self.cached_split_result.update()
        members = set(self.cached_split_result.split_result['cur_eqs'].members(supervoxel_id))
        component.supervoxels = set(x for x in component.supervoxels if x not in members)
        self.state.unused_supervoxels.update(members)
        self._update_view()

    def _exclude_all_but_component(self, s):
        if self.state.selected_component is None:
            return

        component = self.state.components[self.state.selected_component]
        supervoxel_id = self._get_mouse_supervoxel(s)

        if supervoxel_id is None:
            return

        self.cached_split_result.update()
        members = set(self.cached_split_result.split_result['cur_eqs'].members(supervoxel_id))
        new_unused = set(x for x in component.supervoxels if x not in members)
        component.supervoxels = members
        self.state.unused_supervoxels.update(new_unused)
        self._update_view()


    def _make_new_component(self, s):
        del s
        self.state.make_new_component()
        self._update_view()

    def _next_component(self, s):
        del s
        self.state.cycle_selected_component(1)
        self._update_view()

    def _prev_component(self, s):
        del s
        self.state.cycle_selected_component(-1)
        self._update_view()

    def _handle_state_changed(self):
        if self.state.selected_component is None:
            return
        seeds = _get_viewer_seeds(self.viewer.state)
        component = self.state.components[self.state.selected_component]
        if seeds == component.seeds:
            return
        component.seeds = seeds
        with self.viewer.txn() as s:
            self._update_state(s)

    def _get_mouse_supervoxel(self, s):
        supervoxel_id = s.selected_values.get('original')
        if supervoxel_id is not None:
            supervoxel_id = supervoxel_id.value
        if supervoxel_id is None:
            m = s.selected_values.get('split-result')
            if m is not None:
                m = m.value
            if m is not None:
                if isinstance(m, neuroglancer.SegmentIdMapEntry):
                    supervoxel_id = m.key
                else:
                    supervoxel_id = m
        if supervoxel_id is None or supervoxel_id == 0:
            return None
        return supervoxel_id

    def _add_seed(self, s, inclusive):
        supervoxel_id = self._get_mouse_supervoxel(s)
        mouse_voxel_coordinates = s.mouse_voxel_coordinates
        if mouse_voxel_coordinates is None or supervoxel_id is None:
            return
        position = tuple(int(x) for x in mouse_voxel_coordinates)
        self.state.add_seed(supervoxel_id, position, inclusive)
        self._update_view()

    def _update_view(self):
        with self.viewer.txn() as s:
            self._update_state(s)
        with self.viewer.config_state.txn() as s:
            self._update_config_state(s)

    def _update_config_state(self, s):
        if self.state.selected_component is None:
            msg = '[No component selected] %d unused supervoxels' % len(
                self.state.unused_supervoxels)
        else:
            selected_component = self.state.selected_component

            msg = '[Component %d/%d] : %d supervoxels, %d connected components, %d unused' % (
                selected_component, len(self.state.components),
                len(self.cached_split_result.supervoxels),
                len(self.cached_split_result.split_result['cur_eqs'].sets()), len(self.state.unused_supervoxels))
        s.status_messages['status'] = msg

    def _update_state(self, s):
        self.cached_split_result.update()
        self.state.save()
        _set_viewer_seeds(s, self.cached_split_result.seeds)

        s.layers['unused'].segments = self.state.unused_supervoxels
        s.layers['original'].segments = self.cached_split_result.supervoxels
        s.layers['split-result'].segments = self.cached_split_result.supervoxels
        split_result = self.cached_split_result.split_result
        if split_result is not None:
            self._show_split_result(
                s,
                cur_eqs=split_result['cur_eqs'],
            )
        s.layout = neuroglancer.row_layout([
            neuroglancer.LayerGroupViewer(
                layout='3d',
                layers=['image', 'original', 'unused', 'inclusive-seeds', 'exclusive-seeds']),
            neuroglancer.LayerGroupViewer(
                layout='3d', layers=['image', 'split-result', 'inclusive-seeds',
                                     'exclusive-seeds']),
        ])

    def _show_split_result(self, s, cur_eqs):
        split_layer = s.layers['split-result']
        split_layer.equivalences = cur_eqs
        split_layer.segments = set(cur_eqs[x] for x in self.cached_split_result.supervoxels)


def run_batch(args, graph):
    for path in args.split_seeds:
        split_seeds = load_split_seeds(path)
        split_result = do_split(graph=graph, split_seeds=split_seeds, agglo_id=args.agglo_id)
        state = display_split_result(
            graph=graph,
            split_seeds=split_seeds,
            image_url=args.image_url,
            segmentation_url=args.segmentation_url,
            **split_result)
        print('<p><a href="%s">%s</a></p>' % (neuroglancer.to_url(state), path))


def run_interactive(args, graph):
    # Make splitter a global variable so that it is accessible from the
    # interactive `python -i` shell.
    global splitter

    if args.bind_address:
        neuroglancer.set_server_bind_address(args.bind_address)
    if args.static_content_url:
        neuroglancer.set_static_content_source(url=args.static_content_url)

    splitter = InteractiveSplitter(
        graph,
        agglo_id=args.agglo_id,
        image_url=args.image_url,
        segmentation_url=args.segmentation_url,
        state_path=args.state)
    print(splitter.viewer)


def open_graph(path, agglo_id):
    # Check if graph_db is sharded
    graph_db = path
    m = re.match(r'(.*)@([0-9]+)((?:\..*)?)$', graph_db)
    if m is not None:
        num_shards = int(m.group(2))
        shard = agglo_id % num_shards
        graph_db = m.group(1) + ('-%05d-of-%05d' % (shard, num_shards)) + m.group(3)

    return AgglomerationGraph(sqlite3.connect(graph_db, check_same_thread=False))


if __name__ == '__main__':
    ap = argparse.ArgumentParser()

    ap.add_argument('-v', '--verbose', action='store_true', help='Display verbose log messages.')

    common_ap = argparse.ArgumentParser(add_help=False)
    common_ap.add_argument(
        '--graph-db', required=True, help='Path to sqlite3 database specifying agglomeration graph')
    common_ap.add_argument(
        '--image-url', required=True, help='Neuroglancer data source URL for image')
    common_ap.add_argument(
        '--segmentation-url', required=True, help='Neuroglancer data source URL for segmentation')

    sub_aps = ap.add_subparsers(help='command to run')
    interactive_ap = sub_aps.add_parser(
        'interactive', help='Interactively split an aglomerated component', parents=[common_ap])

    batch_ap = sub_aps.add_parser(
        'batch', help='Split based on pre-specified seed files', parents=[common_ap])

    interactive_ap.add_argument(
        '--agglo-id', type=int, required=True, help='Agglomerated component id to split')
    interactive_ap.add_argument('--split-seeds', help='Path to JSON file specifying split seeds')
    interactive_ap.add_argument('--state', help='Path to JSON state file.')
    interactive_ap.add_argument(
        '-a',
        '--bind-address',
        help='Bind address for Python web server.  Use 127.0.0.1 (the default) to restrict access '
        'to browers running on the local machine, use 0.0.0.0 to permit access from remote browsers.'
    )
    interactive_ap.add_argument(
        '--static-content-url', help='Obtain the Neuroglancer client code from the specified URL.')

    interactive_ap.set_defaults(func=run_interactive)

    batch_ap.add_argument(
        '--split-seeds', nargs='+', help='Path to JSON file specifying split seeds')
    batch_ap.add_argument('--agglo-id', type=int, help='Agglomerated component id to split')
    batch_ap.set_defaults(func=run_batch)

    args = ap.parse_args()

    graph = open_graph(args.graph_db, args.agglo_id)

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    args.func(args, graph)
